@@ -40,6 +40,7 @@ from .config import (  # noqa
     get_previous_experiment,
     get_ws_url,
 )
+from .confusion_matrix import ConfusionMatrix
 from .connection import (
     INITIAL_BEAT_DURATION,
     RestServerConnection,
@@ -65,14 +66,7 @@ from .monkey_patching import CometModuleFinder
 from .offline import OfflineExperiment
 from .optimizer import Optimizer
 from .rpc import create_remote_call, get_remote_action_definition
-from .utils import (
-    ConfusionMatrix,
-    Embedding,
-    Histogram,
-    get_comet_version,
-    merge_url,
-    valid_ui_tabs,
-)
+from .utils import Embedding, Histogram, get_comet_version, merge_url, valid_ui_tabs
 
 __author__ = "Gideon<Gideon@comet.ml>"
 __all__ = [
