@@ -46,7 +46,7 @@ class Predictor(object):
         start_step=10,
         time_interval=300,
         n_init_experiments=0,
-        **defaults
+        advanced_config=None,
     ):
         """
         Please email lcp-beta@comet.ml for comments or questions.
@@ -70,8 +70,11 @@ class Predictor(object):
             "HP_batch_size": float("nan"),
             "HP_curr_step": float("nan"),
         }
+
+        if advanced_config is not None:
+            self.set_defaults(**advanced_config)
+
         self.losses = []  # type: List[float]
-        self.set_defaults(**defaults)
         self.threshold = threshold
         if not 0.0 < self.threshold <= 1.0:
             raise ValueError("Threshold must be set between 0.0 and 1.0")
