@@ -23,11 +23,9 @@ import math
 import numbers
 import os
 import os.path
-import random
 import re
 import tempfile
 import time
-from collections import Mapping, defaultdict
 from datetime import datetime
 
 import six
@@ -40,15 +38,20 @@ from ._typing import (
     Callable,
     Dict,
     Generator,
-    Iterable,
     List,
     Optional,
     Sequence,
     Set,
     Tuple,
-    Type,
     Union,
 )
+from .convert_utils import convert_tensor_to_numpy
+
+if six.PY2:
+    from collections import Mapping
+else:
+    from collections.abc import Mapping
+
 
 LOGGER = logging.getLogger(__name__)
 LOG_ONCE_CACHE = set()  # type: Set[str]
@@ -73,6 +76,7 @@ try:
 except ImportError:
     LOGGER.warning("numpy not installed; some functionality will be unavailable")
     HAS_NUMPY = False
+    numpy = None
 
 
 INFINITY = float("inf")
@@ -86,8 +90,7 @@ def fix_special_floats(value, _inf=INFINITY, _neginf=-INFINITY):
     """
 
     try:
-        if hasattr(value, "numpy"):
-            value = value.numpy()
+        value = convert_tensor_to_numpy(value)
 
         # Check if the value is Nan, equivalent of math.isnan
         if math.isnan(value):
@@ -236,7 +239,7 @@ def image_data_to_file_like_object(
     ## Conversion from standard objects to image
     ## Allow file-like objects, numpy arrays, etc.
     if hasattr(image_data, "numpy"):  # pytorch tensor
-        array = image_data.numpy()
+        array = convert_tensor_to_numpy(image_data)
         fp = array_to_image_fp(
             array,
             image_format,
@@ -321,6 +324,8 @@ def array_to_image_fp(
     image = array_to_image(
         array, image_scale, image_shape, image_colormap, image_minmax, image_channels
     )
+    if not image:
+        return None
     return image_to_fp(image, image_format)
 
 
@@ -343,7 +348,7 @@ def array_to_image(
         from matplotlib import cm
     except ImportError:
         LOGGER.error(
-            "The Python libraries PIL, numpy, and matplotlib are required for this operation"
+            "The Python libraries PIL, numpy, and matplotlib are required for converting a numpy array into an image"
         )
         return None
 
@@ -715,597 +720,6 @@ class Embedding(object):
         return template
 
 
-class ConfusionMatrix(object):
-    """
-    Data structure for holding a confusion matrix of values and their
-    labels.
-    """
-
-    def __init__(
-        self,
-        y_true=None,
-        y_predicted=None,
-        labels=None,
-        matrix=None,
-        title="Confusion Matrix",
-        row_label="Actual Category",
-        column_label="Predicted Category",
-        max_examples_per_cell=25,
-        max_categories=25,
-        winner_function=None,
-        index_to_example_function=None,
-        cache=True,
-        selected=None,
-        **kwargs  # keyword args for index_to_example_function
-    ):
-        """
-        Create the ConfusionMatrix data structure.
-
-        Args:
-            y_true: (optional) a list of target vectors containing
-                the correct values for the y_predicted vectors. If
-                not provided, then matrix may be provided.
-            y_predicted: (optional) a list of output vectors containing
-                the predicted values for the y_true vectors. If
-                not provided, then matrix may be provided.
-            labels: (optional) a list of strings that name of the
-                columns and rows, in order. By default, it will be
-                "0" through the number of categories (e.g., rows/columns).
-            matrix: (optional) the confusion matrix (list of lists).
-                Must be square, if given. If not given, then it is
-                possible to provide y_true and y_predicted.
-            title: (optional) a custom name to be displayed. By
-                default, it is "Confusion Matrix".
-            row_label: (optional) label for rows. By default, it is
-                "Actual Category".
-            column_label: (optional) label for columns. By default,
-                it is "Predicted Category".
-            max_examples_per_cell: (optional) maximum number of
-                examples per cell. By default, it is 25.
-            max_categories: (optional) max number of columns and rows to
-                use. By default, it is 25.
-            winner_function: (optional) a function that takes in an
-                entire list of rows of patterns, and returns
-                the winning category for each row. By default, it is argmax.
-            index_to_example_function: (optional) a function
-                that takes an index and returns either
-                a number, a string, a URL, or a {"sample": str,
-                "assetId": str} dictionary. See below for more info.
-                By default, the function returns a number representing
-                the index of the example.
-            cache: (optional) should the results of index_to_example_function
-                be cached and reused? By default, cache is True.
-            selected: (optional) None, or list of selected category
-                indices. These are the rows/columns that will be shown. By
-                default, select is None. If the number of categories is
-                greater than max_categories, and selected is not provided,
-                then selected will be computed automatically by selecting
-                the most confused categories.
-            kwargs: (optional) any extra keywords and their values will
-                be passed onto the index_to_example_function.
-
-        Note:
-            The matrix is [row][col] and [real][predicted] order. That way, the
-            data appears as it is display in the confusion matrix in the user
-            interface on comet.ml.
-
-        Example:
-
-        ```python
-        # Typically, you can log the y_true/y_predicted or matrix:
-
-        >>> experiment = Experiment()
-
-        # If you have a y_true and y_predicted:
-        >>> y_predicted = model.predict(x_test)
-        >>> experiment.log_confusion_matrix(y_true, y_predicted)
-
-        # Or, if you have already computed the matrix:
-        >>> experiment.log_confusion_matrix(labels=["one", two", three"],
-                                            matrix=[[10, 0, 0]
-                                                    [ 0, 9, 1]
-                                                    [ 1, 1, 8]])
-
-        # However, if you want to reuse examples from previous runs,
-        # you can reuse a ConfusionMatrix instance. You might want to
-        # do this if you are creating a series of confusion matrices
-        # during the training of a model.
-        # See https://staging.comet.ml/docs/quick-start/ for a tutorial.
-
-        >>> cm = ConfusionMatrix()
-        >>> y_predicted = model.predict(x_test)
-        >>> cm.compute_matrix(y_true, y_predicted)
-        >>> experiment.log_confusion_matrix(matrix=cm)
-
-        # Log again, using previously cached values:
-        >>> y_predicted = model.predict(x_test)
-        >>> cm.compute_matrix(y_true, y_predicted)
-        >>> experiment.log_confusion_matrix(matrix=cm)
-        ```
-        """
-        if y_true is not None and y_predicted is not None:
-            if matrix is not None:
-                raise ValueError(
-                    "you need to give either (y_true and y_predicted) or matrix, NOT both"
-                )
-            # else fine
-        elif y_true is None and y_predicted is None:
-            pass  # fine
-        elif y_true is None or y_predicted is None:
-            raise ValueError("if you give y_true OR y_predicted you must give both")
-
-        if winner_function is not None:
-            self.winner_function = winner_function
-        else:
-            self.winner_function = self.default_winner_function
-
-        if index_to_example_function is not None:
-            self.index_to_example_function = index_to_example_function
-        else:
-            self.index_to_example_function = self.default_index_to_example_function
-
-        self.y_true = y_true
-        self.y_predicted = y_predicted
-        self.labels = labels
-        self.title = title
-        self.row_label = row_label
-        self.column_label = column_label
-        self.max_examples_per_cell = max_examples_per_cell
-        self.max_categories = max_categories
-        self.selected = selected
-        self.use_cache = cache
-        self.clear_cache()
-        self.clear()
-        self._matrix = None
-
-        if self.y_true is not None and self.y_predicted is not None:
-            self.compute_matrix(self.y_true, self.y_predicted, **kwargs)
-        elif matrix is not None:
-            try:
-                self._matrix = convert_to_matrix(matrix, dtype=numbers.Number)
-            except Exception:
-                LOGGER.error(
-                    "convert_to_matrix failed; confusion matrix not generated",
-                    exc_info=True,
-                )
-
-    def clear(self):
-        """
-        Clear the matrices and type.
-        """
-        self.type = None
-        self._example_matrix = None
-        self._matrix = None
-
-    def clear_cache(self):
-        """
-        Clear the caches.
-        """
-        # Set of indices (ints):
-        self._cache = set()  # type: Set[int]
-        # Map index (int) -> example
-        self._cache_example = {}  # type: Dict[int, Any]
-
-    def initialize(self):
-        """
-        Initialize the confusion matrix based on y_true.
-        """
-        n = len(self.y_true[0])
-        self._matrix = self._create_matrix((n, n), 0)
-        self._example_matrix = self._create_matrix((n, n), None)
-
-    def default_winner_function(self, ndarray):
-        """
-        A default winner function. Takes a list
-        of patterns to apply winner function to.
-
-        Args:
-            ndarry: a 2-D matrix where rows are the patterns
-
-        Returns a list of winning categories.
-        """
-        if HAS_NUMPY:
-
-            def winner(ndarray):
-                return numpy.argmax(ndarray, axis=1)
-
-        else:
-            # numpy is faster, but not required
-            log_once_at_level(
-                logging.INFO,
-                "numpy not installed; using a slower "
-                + "winner_function for confusion matrix",
-            )
-
-            def winner(ndarray):
-                # Even if the following code is doing two iterations on the
-                # list, most of the computation is done by C code
-                return [array.index(max(array)) for array in ndarray]
-
-        return winner(ndarray)
-
-    def default_index_to_example_function(self, index, **kwargs):
-        """
-        User-provided function.
-
-        Args:
-            index: the index of the pattern being tested
-            kwargs: additional keyword arguments for an overridden method
-
-        Returns:
-            * an integer representing the winning cateory
-            * a string representing a example
-            * a string representing a URL (starts with "http")
-            * a dictionary containing keys "sample" and "assetId"
-
-        The return dictionary is used to link a confusion matrix cell
-        with a Comet asset. In this function, you can create an asset
-        and return a dictionary, like so:
-
-        ```python
-        # Example index_to_example_function
-        def index_to_example_function(index):
-            # x_test is user's inputs (just an example):
-            image_array = x_test[index]
-            # Make an asset name:
-            image_name = "confusion-matrix-%05d.png" % index
-            # Make an asset:
-            results = experiment.log_image(
-                image_array, name=image_name, image_shape=(28, 28, 1)
-            )
-            # Return the example name and assetId
-            return {"sample": image_name, "assetId": results["imageId"]}
-
-        # Then, pass it to ConfusionMatrix(), or log_confusion_matrix()
-        ```
-        """
-        return index
-
-    def _set_type_from_example(self, example):
-        """
-        Take the cached example and set the global
-        confusion matrix type.
-
-        Args:
-            example: an int or dict
-        """
-        if isinstance(example, int):
-            self.type = "integer"
-        elif isinstance(example, dict):
-            if example["assetId"] is not None:
-                self.type = "image"
-            elif example["sample"].startswith("http"):
-                self.type = "link"
-            else:
-                self.type = "string"
-        else:
-            raise TypeError("unknown example type: %r" % example)
-
-    def _process_new_example(self, example, index):
-        """
-        Turn the user's return value into a proper example.  Sets the type
-        based on user's value.
-
-        Args:
-            example: a new example from user function
-            index: the index of the example
-
-        Side-effect: saves in cache if possible.
-        """
-        if isinstance(example, int):
-            self.type = "integer"
-        elif isinstance(example, str):
-            if example.startswith("http"):
-                self.type = "link"
-            else:
-                self.type = "string"
-            example = {
-                "index": index,  # index
-                "sample": example,  # example
-                "assetId": None,  # assetId
-            }
-        elif isinstance(example, dict):
-            # a dict of index (int), assetId (string), example (string)
-            if "sample" not in example or "assetId" not in example:
-                raise ValueError(
-                    "index_to_example_function must return {'sample': ..., 'assetId': ...}"
-                )
-            # Add the index, in case not already in:
-            if "index" not in example:
-                example["index"] = index
-            # Set the confusion matrix type:
-            if "type" in example:
-                self.type = example["type"]
-                # Remove from dict:
-                del example["type"]
-            else:  # default type
-                self.type = "image"
-        else:
-            raise ValueError(
-                "index_to_example_function must return an int, string, URL, or {'sample': string, 'assetId': string}, not %r"
-                % type(example)
-            )
-        if self.type != "integer" and self.use_cache:
-            self._put_example_in_cache(index, example)
-        return example
-
-    def _index_to_example(self, index, **kwargs):
-        """
-        Wrapper around user function/cache.
-
-        Args:
-            index: the index of the example
-            kwargs: passed to user function
-        """
-        if self.use_cache and self._example_in_cache(index):
-            example = self._get_example_from_cache(index)
-            self._set_type_from_example(example)
-            return example
-
-        try:
-            example = self.index_to_example_function(index, **kwargs)
-        except Exception:
-            LOGGER.error(
-                "The index_to_example_function failed for index %s; example not generated",
-                index,
-                exc_info=True,
-                extra={"show_traceback": True},
-            )
-            example = index
-        example = self._process_new_example(example, index)
-
-        return example
-
-    def _create_matrix(self, shape, initial_value):
-        # type: (Tuple[int, int], Any) -> List[List[Any]]
-        """
-        Create a matrix with initial value of initial_value.
-
-        Args:
-            shape: shape of matrix (row, col)
-            initial_value: initial value for cell
-        """
-        return [[initial_value for x in range(shape[1])] for y in range(shape[0])]
-
-    def _get_example_from_cache(self, index):
-        """
-        Get a example from the example cache.
-
-        Args:
-            index: the index of example
-        """
-        key = index
-        return self._cache_example[key]
-
-    def _example_in_cache(self, index):
-        """
-        Is the example in the example cache?
-
-        Args:
-            index: the index of example
-        """
-        key = index
-        return key in self._cache_example
-
-    def _put_example_in_cache(self, index, example):
-        """
-        Put a example in the example cache.
-
-        Args:
-            index: the index of example
-            example: the processed example
-        """
-        key = index
-        self._cache_example[key] = example
-
-    def _example_from_list(self, indices, x, y, **kwargs):
-        # type: (Set[int], int, int, Any) -> Set[int]
-        """
-        Example from indices so that it is no more than max length.
-        Use previous indices from cache.
-
-        Args:
-            indices: the indices of the patterns to example from
-            x: the column of example cell
-            y: the row of example cell
-            kwargs: keyword args to pass to user function
-        """
-        if len(indices) <= self.max_examples_per_cell:
-            retval = set(indices)
-        else:
-            indices = list(indices)
-            retval = list(self._cache.intersection(indices))
-            # If you need more:
-            retval += [
-                indices.pop(random.randint(0, len(indices) - 1))
-                for i in range(self.max_examples_per_cell - len(retval))
-            ]
-            # Return minimum needed:
-            retval = set(retval[: self.max_examples_per_cell])
-
-        new_ones = retval - self._cache
-        if self.index_to_example_function is not None:
-            examples = []
-            for index in retval:
-                example = self._index_to_example(index, **kwargs)
-                examples.append(example)
-            self._example_matrix[x][y] = examples
-        # Update the ones sent:
-        if self.use_cache:
-            self._cache.update(new_ones)
-        return retval
-
-    def compute_matrix(
-        self, y_true, y_predicted, index_to_example_function=None, **kwargs
-    ):
-        """
-        Compute the confusion matrix.
-
-        Args:
-            y_true: list of vectors representing the targets
-            y_predicted: list of vectors representing predicted
-                values
-            index_to_example_function: (optional) a function
-                that takes an index and returns either
-                a number, a string, a URL, or a {"sample": str,
-                "assetId": str} dictionary. See below for more info.
-                By default, the function returns a number representing
-                the index of the example.
-
-        Note: uses winner_function to compute winning category.
-        """
-        if len(y_true) != len(y_predicted):
-            raise ValueError(
-                "y_true and y_predicted should have the same lengths; %s != %s"
-                % (len(y_true), len(y_predicted))
-            )
-        if len(y_true[0]) != len(y_predicted[0]):
-            raise ValueError(
-                "y_true[0] and y_predicted[0] should have the same lengths; %s != %s"
-                % (len(y_true[0]), len(y_predicted[0]))
-            )
-        self.y_true = y_true
-        self.y_predicted = y_predicted
-        if index_to_example_function is not None:
-            self.index_to_example_function = index_to_example_function
-        self.initialize()
-
-        # Keep track of all indices for each cell:
-        # Create confusion matrix:
-        xs = self.winner_function(y_true)  # type: Iterable[Any]
-        ys = self.winner_function(y_predicted)  # type: Iterable[Any]
-
-        examples = defaultdict(set)  # type: Dict[Tuple[int, int], Set[int]]
-        for (i, (raw_x, raw_y)) in enumerate(zip(xs, ys)):
-            try:
-                x = convert_to_scalar(raw_x, dtype=int)  # type: int
-            except TypeError:
-                LOGGER.warning("Invalid y_true value %r, ignoring it", raw_x)
-                continue
-
-            try:
-                y = convert_to_scalar(raw_y, dtype=int)  # type: int
-            except TypeError:
-                LOGGER.warning("Invalid y_predictor value %r, ignoring it", raw_y)
-                continue
-
-            # Add count to cell:
-            self._matrix[x][y] += 1
-            # Add index to cell:
-            examples[(x, y)].add(i)
-
-        # Example all cells that have items (reuse from cache/other cells if possible):
-        for key in examples:
-            x, y = key
-            if self.selected is not None:
-                if x in self.selected and y in self.selected:
-                    self._example_from_list(examples[key], x, y, **kwargs)
-            else:
-                self._example_from_list(examples[key], x, y, **kwargs)
-
-    def to_json(self):
-        """
-        Return the associated confusion matrix as the JSON to
-        upload.
-        """
-        if (
-            self.y_true is not None
-            and len(self.y_true[0]) > self.max_categories
-            and self.selected is None
-        ):
-            # Sort row by worst performing:
-            correct_counts = [(i, self._matrix[i][i]) for i in range(len(self._matrix))]
-            ordered_rows = sorted(correct_counts, key=lambda pair: pair[1])
-            self.selected = [row[0] for row in ordered_rows[: self.max_categories]]
-
-        if self.selected is not None:
-            # Make sure sorted, smallest to largest:
-            self.selected = sorted(self.selected)
-
-        if self.labels is None:
-            if self.selected is not None:
-                labels = [str(v) for v in self.selected]
-            elif self._matrix is not None:
-                labels = [str(v) for v in range(len(self._matrix[0]))]
-            else:
-                labels = []
-        else:
-            if self.selected is not None:
-                labels = [
-                    label for (i, label) in enumerate(self.labels) if i in self.selected
-                ]
-            else:
-                labels = self.labels
-        if self._example_matrix is not None:
-            smatrix = self._example_matrix if any(self._example_matrix) else None
-            if smatrix is not None and self.selected is not None:
-                smatrix = [
-                    [smatrix[row][col] for col in self.selected]
-                    for row in self.selected
-                ]
-        else:
-            smatrix = None
-        if smatrix is None:
-            self.type = None
-        if self.selected is not None:
-            matrix = [
-                [self._matrix[row][col] for col in self.selected]
-                for row in self.selected
-            ]
-        else:
-            matrix = self._matrix
-        if matrix is not None:
-            if len(matrix) != len(labels):
-                raise ValueError(
-                    "The length of labels does not match number of categories"
-                )
-        return {
-            "version": 1,
-            "title": self.title,
-            "labels": labels,
-            "matrix": matrix,
-            "rowLabel": self.row_label,
-            "columnLabel": self.column_label,
-            "maxSamplesPerCell": self.max_examples_per_cell,
-            "sampleMatrix": smatrix,
-            "type": self.type,
-        }
-
-    def display(self, space=4):
-        """
-        Display an ASCII version of the confusion matrix.
-
-        Args:
-            space: (optional) column width
-        """
-
-        def format(string):
-            print(("%" + str(space) + "s") % str(string)[: space - 1], end="")
-
-        json_format = self.to_json()
-        total_width = len(json_format["matrix"]) * space
-        row_label = json_format["rowLabel"] + (" " * total_width)
-        format(row_label[0])
-        format("")
-        print(json_format["title"].center(total_width))
-        format(row_label[1])
-        format("")
-        print(json_format["columnLabel"].center(total_width))
-        format(row_label[2])
-        format("")
-        for row in range(len(json_format["matrix"])):
-            format(json_format["labels"][row])
-        print()
-        format(row_label[3])
-        for row in range(len(json_format["matrix"])):
-            format(json_format["labels"][row])
-            for col in range(len(json_format["matrix"][row])):
-                format(json_format["matrix"][row][col])
-            print()
-            format(row_label[row + 4])
-        print()
-
-
 class Histogram(object):
     """
     Data structure for holding a counts of values. Creates an
@@ -2012,27 +1426,6 @@ def valid_ui_tabs(tab=None, preferred=False):
         raise ValueError("invalid tab name; tab should be in %r" % preferred_names)
 
 
-def convert_to_matrix(matrix, dtype=None):
-    """
-    Convert an unknown item into a list of lists of scalars
-    and ensure type is dtype (if given).
-    """
-    # First, convert it to numpy if possible:
-    if hasattr(matrix, "numpy"):  # pytorch tensor
-        matrix = matrix.numpy()
-    elif hasattr(matrix, "eval"):  # tensorflow tensor
-        matrix = matrix.eval()
-
-    # Next, convert to lists of scalars:
-    if hasattr(matrix, "tolist"):  # numpy array
-        if len(matrix.shape) != 2:
-            raise ValueError("matrix should be two dimensional")
-        return matrix.tolist()
-    else:
-        # assume it is something we can iterate over:
-        return [convert_to_list(row, dtype=dtype) for row in matrix]
-
-
 def shape(data):
     """
     Given a nested list or a numpy array,
@@ -2090,6 +1483,11 @@ def fast_flatten(items):
     if isinstance(items, (numbers.Number, six.string_types)):
         return items
 
+    try:
+        items = convert_tensor_to_numpy(items)
+    except Exception:
+        LOGGER.debug("unable to convert tensor; continuing", exc_info=True)
+
     if HAS_NUMPY:
         try:
             # Vector, Matrix conversion:
@@ -2114,27 +1512,6 @@ def fast_flatten(items):
         return flatten(items)
 
 
-def convert_to_list(items, dtype=None):
-    """
-    Take an unknown item and convert to a list of scalars
-    and ensure type is dtype, if given.
-    """
-    # First, convert it to numpy if possible:
-    if hasattr(items, "numpy"):  # pytorch tensor
-        items = items.numpy()
-    elif hasattr(items, "eval"):  # tensorflow tensor
-        items = items.eval()
-
-    # Next, handle numpy array:
-    if hasattr(items, "tolist"):
-        if len(items.shape) != 1:
-            raise ValueError("list should be one dimensional")
-        return items.tolist()
-    else:
-        # assume it is something with numbers in it:
-        return [convert_to_scalar(item, dtype=dtype) for item in items]
-
-
 def convert_to_string(user_data):
     # type: (Any) -> str
     """
@@ -2144,7 +1521,7 @@ def convert_to_string(user_data):
         user_data = json.dumps(user_data, sort_keys=True)
 
     if hasattr(user_data, "numpy"):
-        user_data = user_data.numpy()
+        user_data = convert_tensor_to_numpy(user_data)
 
     if isinstance(user_data, bytes) and not isinstance(user_data, str):
         user_data = user_data.decode("utf-8")
@@ -2175,41 +1552,6 @@ def convert_to_string_value(user_data):
     return convert_to_string_truncated(user_data, 1000)
 
 
-def convert_to_scalar(user_data, dtype=None):
-    # type: (Any, Optional[Type]) -> Any
-    """
-    Try to ensure that the given user_data is converted back to a
-    Python scalar, and of proper type (if given).
-    """
-
-    # First try to convert tensorflow tensor to numpy objects
-    try:
-        if hasattr(user_data, "numpy"):
-            user_data = user_data.numpy()
-    except Exception:
-        LOGGER.debug(
-            "Failed to convert tensorflow tensor %r to numpy object",
-            user_data,
-            exc_info=True,
-        )
-
-    # Then try to convert numpy object to a Python scalar
-    try:
-        if hasattr(user_data, "item") and callable(user_data.item):
-            user_data = user_data.item()
-    except Exception:
-        LOGGER.debug(
-            "Failed to convert numpy object %r to Python scalar",
-            user_data,
-            exc_info=True,
-        )
-
-    if dtype is not None and not isinstance(user_data, dtype):
-        raise TypeError("%r is not of type %r" % (user_data, dtype))
-
-    return user_data
-
-
 def makedirs(name, exist_ok=False):
     """
     Replacement for Python2's version lacking exist_ok
